<?php

namespace Slack\Blockkit;

use Illuminate\Support\Collection;
use Slack\BlockKit;

/**
 * Class Blockkit Block
 * Represents a Block used in Blockkit
 *
 * @package Slack\Blockkit
 */
class Block extends BlockKit
{
	/**
	 * Add Actions Block
	 *
	 * @param Collection $elements
	 * @return $this
	 */
	public function addAction(Collection $elements): self
	{
		// Initialise
		$this->_data = collect();

		$this->_data->put('type','actions');
		$this->_data->put('elements',$elements);

		return $this;
	}

	/**
	 * A context block
	 *
	 * @param Collection $elements
	 * @return $this
	 */
	public function addContext(Collection $elements): self
	{
		// Initialise
		$this->_data = collect();

		$this->_data->put('type','context');
		$this->_data->put('elements',$elements);

		return $this;
	}

	/**
	 * Add a bock divider
	 */
	public function addDivider(): self
	{
		$this->_data->put('type','divider');

		return $this;
	}

	/**
	 * Add a block header
	 *
	 * @param string $text
	 * @param string $type
	 * @return Block
	 * @throws \Exception
	 */
	public function addHeader(string $text,string $type='plain_text'): self
	{
		$this->_data->put('type','header');
		$this->_data->put('text',$this->text($text,$type));

		return $this;
	}

	/**
	 * Generates a multiselect that queries back to the server for values
	 *
	 * @param string $label
	 * @param string $action
	 * @return $this
	 * @throws \Exception
	 */
	public function addMultiSelectInput(string $label,string $action): self
	{
		$this->_data->put('type','section');
		$this->_data->put('text',$this->text('mrkdwn',$label));
		$this->_data->put('accessory',[
			'action_id'=>$action,
			'type'=>'multi_external_select',
		]);

		return $this;
	}

	/**
	 * @param string          $label
	 * @param string          $action
	 * @param Collection      $options
	 * @param Collection|null $selected
	 * @param int|null        $maximum
	 * @return $this
	 * @throws \Exception
	 */
	public function addMultiSelectStaticInput(string $label,string $action,Collection $options,Collection $selected=NULL,int $maximum=NULL): self
	{
		$this->_data->put('type','section');
		$this->_data->put('text',$this->text($label,'mrkdwn'));

		$x = collect();
		$x->put('action_id',$action);
		$x->put('type','multi_static_select');
		$x->put('options',$options->transform(function ($item) {
			return ['text'=>$this->text($item->name,'plain_text'),'value'=>(string)$item->id];
		}));

		if ($selected and $selected->count())
			$x->put('initial_options',$selected->transform(function ($item) {
				return ['text'=>$this->text($item->name,'plain_text'),'value'=>(string)$item->id];
			}));

		if ($maximum)
			$x->put('max_selected_items',$maximum);

		$this->_data->put('accessory',$x);

		return $this;
	}

	/**
	 * @param Collection $options
	 * @param string     $action
	 * @return Collection
	 */
	public function addOverflow(Collection $options,string $action): Collection
	{
		return collect([
			'type'=>'overflow',
			'options'=>$options,
			'action_id'=>$action,
		]);
	}

	/**
	 * A section block
	 *
	 * @param string          $text
	 * @param string          $type
	 * @param Collection|null $accessories
	 * @param string|null     $block_id
	 * @return $this
	 * @throws \Exception
	 */
	public function addSection(string $text,string $type='mrkdwn',Collection $accessories=NULL,string $block_id=NULL): self
	{
		// Initialise
		$this->_data = collect();

		$this->_data->put('type','section');
		$this->_data->put('text',$this->text($text,$type));

		if ($block_id)
			$this->_data->put('block_id',$block_id);

		if ($accessories AND $accessories->count())
			$this->_data->put('accessory',$accessories);

		return $this;
	}

	/**
	 * @param string      $label
	 * @param string      $action
	 * @param Collection  $options
	 * @param string|null $default
	 * @return $this
	 * @throws \Exception
	 */
	public function addSelect(string $label,string $action,Collection $options,string $default=NULL): self
	{
		$this->_data->put('type','section');
		$this->_data->put('text',$this->text($label,'mrkdwn'));

		// Accessories
		$x = collect();
		$x->put('action_id',$action);
		$x->put('type','static_select');
		$x->put('options',$options->map(function ($item) {
			if (is_array($item))
				$item = (object)$item;

			return [
				'text'=>[
					'type'=>'plain_text',
					'text'=>(string)$item->name,
				],
				'value'=>(string)($item->value ?: $item->id)
			];
		}));

		if ($default) {
			$choice = $options->filter(function($item) use ($default) {
				if (is_array($item))
					$item = (object)$item;

				return ($item->value == $default) ? $item : NULL;
			})->filter()->pop();

			if ($choice) {
				$x->put('initial_option',[
					'text'=>$this->text($choice['name'],'plain_text'),
					'value'=>(string)$choice['value']
				]);
			}
		}

		$this->_data->put('accessory',$x);

		return $this;
	}

	/**
	 * Generates a single-line input dialog
	 *
	 * @param string $label
	 * @param string $action
	 * @param string $placeholder
	 * @param int    $minlength
	 * @param string $hint
	 * @param string $initial
	 * @return $this
	 * @throws \Exception
	 */
	public function addSingleLineInput(string $label,string $action,string $placeholder='',int $minlength=5,string $hint='',string $initial=''): self
	{
		return $this->input($label,$action,$minlength,$placeholder,FALSE,$hint,$initial);
	}

	/**
	 * Generates a multi-line input dialog
	 *
	 * @param string $label
	 * @param string $action
	 * @param string $placeholder
	 * @param int    $minlength
	 * @param string $hint
	 * @param string $initial
	 * @return $this
	 * @throws \Exception
	 */
	public function addMultiLineInput(string $label,string $action,string $placeholder='',int $minlength=20,string $hint='',string $initial=''): self
	{
		return $this->input($label,$action,$minlength,$placeholder,TRUE,$hint,$initial);
	}

}
